from __future__ import annotations

import collections
import copy
import functools
import importlib
import inspect
import math
import pickle
import random
import statistics

import numpy as np
import pytest
from scipy import stats as sp_stats

from river import stats, utils


def load_stats():
    for _, obj in inspect.getmembers(importlib.import_module("river.stats"), inspect.isclass):
        try:
            if inspect.isabstract(obj):
                continue

            if issubclass(obj, stats.Link):
                yield obj(stats.Shift(1), stats.Mean())
                continue

            sig = inspect.signature(obj)
            yield obj(
                **{
                    param.name: param.default if param.default != param.empty else 1
                    for param in sig.parameters.values()
                }
            )
        except ValueError:
            yield obj()


@pytest.mark.parametrize("stat", load_stats(), ids=lambda stat: stat.__class__.__name__)
def test_pickling(stat):
    assert isinstance(pickle.loads(pickle.dumps(stat)), stat.__class__)
    assert isinstance(copy.deepcopy(stat), stat.__class__)

    # Check the statistic has a working __str__ and name method
    assert isinstance(str(stat), str)

    if isinstance(stat, stats.base.Univariate):
        assert isinstance(stat.name, str)


@pytest.mark.parametrize("stat", load_stats(), ids=lambda stat: stat.__class__.__name__)
def test_pickling_value(stat):
    for i in range(10):
        if isinstance(stat, stats.base.Bivariate):
            stat.update(i, i)
        elif isinstance(stat, stats.NUnique):  # takes string in input
            stat.update(str(i))
        else:
            stat.update(i)

    assert stat.get() == pickle.loads(pickle.dumps(stat)).get()
    assert stat.get() == copy.deepcopy(stat).get()


@pytest.mark.parametrize("stat", load_stats(), ids=lambda stat: stat.__class__.__name__)
def test_repr_with_no_updates(stat):
    assert isinstance(repr(stat), str)
    assert isinstance(str(stat), str)


@pytest.mark.parametrize(
    "stat, func",
    [
        (stats.Kurtosis(bias=True), sp_stats.kurtosis),
        (stats.Kurtosis(bias=False), functools.partial(sp_stats.kurtosis, bias=False)),
        (stats.Mean(), statistics.mean),
        (stats.Skew(bias=True), sp_stats.skew),
        (stats.Skew(bias=False), functools.partial(sp_stats.skew, bias=False)),
        (stats.Var(ddof=0), np.var),
        (stats.Var(), functools.partial(np.var, ddof=1)),
    ],
)
def test_univariate(stat, func):
    X = [random.random() for _ in range(30)]

    for i, x in enumerate(X):
        stat.update(x)
        if i >= 1:
            assert math.isclose(stat.get(), func(X[: i + 1]), abs_tol=1e-10)


# TODO
# def _weighted_variance(X, W):
#     mean = np.average(X, weights=W)
#     return np.average((W - mean) ** 2, weights=W)


@pytest.mark.parametrize(
    "stat, func",
    [
        (stats.Mean(), lambda x, w: np.average(x, weights=w)),
    ],
)
def test_univariate_frequency_weights(stat, func):
    """https://www.wikiwand.com/en/Weighted_arithmetic_mean"""

    X = [random.random() for _ in range(30)]
    W = [random.randint(1, 5) for _ in range(30)]

    for i, (x, w) in enumerate(zip(X, W)):
        stat.update(x, w)
        if i >= 1:
            assert math.isclose(stat.get(), func(X[: i + 1], W[: i + 1]), abs_tol=1e-10)


@pytest.mark.parametrize(
    "stat, func",
    [
        (stats.Mean(), lambda x, w: np.average(x, weights=w)),
    ],
)
def test_univariate_reliability_weights(stat, func):
    """https://www.wikiwand.com/en/Weighted_arithmetic_mean"""

    X = [random.random() for _ in range(30)]
    W = [random.random() for _ in range(30)]

    for i, (x, w) in enumerate(zip(X, W)):
        stat.update(x, w)
        if i >= 1:
            assert math.isclose(stat.get(), func(X[: i + 1], W[: i + 1]), abs_tol=1e-10)


@pytest.mark.parametrize(
    "stat, func",
    [
        # TODO: we shouldn't ignore these types
        (utils.Rolling(stats.Mean(), 3), statistics.mean),  # type: ignore
        (utils.Rolling(stats.Mean(), 10), statistics.mean),  # type: ignore
        (utils.Rolling(stats.Var(ddof=0), 3), np.var),  # type: ignore
        (utils.Rolling(stats.Var(ddof=0), 10), np.var),  # type: ignore
        (
            stats.RollingQuantile(0.0, 10),
            functools.partial(np.quantile, q=0.0, method="linear"),
        ),
        (
            stats.RollingQuantile(0.25, 10),
            functools.partial(np.quantile, q=0.25, method="linear"),
        ),
        (
            stats.RollingQuantile(0.5, 10),
            functools.partial(np.quantile, q=0.5, method="linear"),
        ),
        (
            stats.RollingQuantile(0.75, 10),
            functools.partial(np.quantile, q=0.75, method="linear"),
        ),
        (
            stats.RollingQuantile(1, 10),
            functools.partial(np.quantile, q=1, method="linear"),
        ),
    ],
)
def test_rolling_univariate(stat, func):
    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    n = stat.window_size
    X = [random.random() for _ in range(30)]

    for i, x in enumerate(X):
        stat.update(x)
        if i >= 1:
            assert math.isclose(stat.get(), func(tail(X[: i + 1], n)), abs_tol=1e-10)


@pytest.mark.parametrize(
    "stat, func",
    [
        # TODO: we shouldn't ignore these types
        (utils.Rolling(stats.Mean(), 3), lambda x, w: np.average(x, weights=w)),  # type: ignore
        (utils.Rolling(stats.Mean(), 10), lambda x, w: np.average(x, weights=w)),  # type: ignore
    ],
)
def test_rolling_univariate_sample_weights(stat, func):
    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    n = stat.window_size
    X = [random.random() for _ in range(30)]
    W = [random.randint(1, 5) for _ in range(30)]

    for i, (x, w) in enumerate(zip(X, W)):
        stat.update(x, w)
        if i >= 1:
            assert math.isclose(
                stat.get(), func(tail(X[: i + 1], n), tail(W[: i + 1], n)), abs_tol=1e-10
            )


@pytest.mark.parametrize(
    "stat, func",
    [
        # TODO: we shouldn't ignore these types
        (utils.Rolling(stats.Mean(), 3), lambda x, w: np.average(x, weights=w)),  # type: ignore
        (utils.Rolling(stats.Mean(), 10), lambda x, w: np.average(x, weights=w)),  # type: ignore
    ],
)
def test_rolling_univariate_reliability_weights(stat, func):
    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    n = stat.window_size
    X = [random.random() for _ in range(30)]
    W = [random.random() for _ in range(30)]

    for i, (x, w) in enumerate(zip(X, W)):
        stat.update(x, w)
        if i >= 1:
            assert math.isclose(
                stat.get(), func(tail(X[: i + 1], n), tail(W[: i + 1], n)), abs_tol=1e-10
            )


@pytest.mark.parametrize(
    "stat, func",
    [
        (stats.Cov(), lambda x, y: np.cov(x, y)[0, 1]),
        (stats.PearsonCorr(), lambda x, y: sp_stats.pearsonr(x, y)[0]),
    ],
)
def test_bivariate(stat, func):
    X = [random.random() for _ in range(30)]
    Y = [random.random() * x for x in X]

    for i, (x, y) in enumerate(zip(X, Y)):
        stat.update(x, y)
        if i >= 1:
            assert math.isclose(stat.get(), func(X[: i + 1], Y[: i + 1]), abs_tol=1e-10)


@pytest.mark.parametrize(
    "stat, func",
    [
        (utils.Rolling(stats.PearsonCorr(), 3), lambda x, y: sp_stats.pearsonr(x, y)[0]),  # type: ignore
        (utils.Rolling(stats.PearsonCorr(), 10), lambda x, y: sp_stats.pearsonr(x, y)[0]),  # type: ignore
        (utils.Rolling(stats.Cov(), 3), lambda x, y: np.cov(x, y)[0, 1]),  # type: ignore
        (utils.Rolling(stats.Cov(), 10), lambda x, y: np.cov(x, y)[0, 1]),  # type: ignore
    ],
)
def test_rolling_bivariate(stat, func):
    # Enough alread

    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    n = stat.window_size
    X = [random.random() for _ in range(30)]
    Y = [random.random() * x for x in X]

    for i, (x, y) in enumerate(zip(X, Y)):
        stat.update(x, y)
        if i >= 1:
            x_tail = tail(X[: i + 1], n)
            y_tail = tail(Y[: i + 1], n)
            assert math.isclose(stat.get(), func(x_tail, y_tail), abs_tol=1e-10)


@pytest.mark.parametrize(
    "stat",
    filter(
        lambda stat: hasattr(stat, "update_many")
        and issubclass(stat.__class__, stats.base.Univariate),
        load_stats(),
    ),
    ids=lambda stat: stat.__class__.__name__,
)
def test_update_many_univariate(stat):
    batch_stat = stat.clone()

    for _ in range(5):
        X = np.random.random(10)
        batch_stat.update_many(X)
        for x in X:
            stat.update(x)

    assert math.isclose(batch_stat.get(), stat.get())


@pytest.mark.parametrize(
    "stat",
    filter(
        lambda stat: hasattr(stat, "update_many")
        and issubclass(stat.__class__, stats.base.Bivariate),
        load_stats(),
    ),
    ids=lambda stat: stat.__class__.__name__,
)
def test_update_many_bivariate(stat):
    batch_stat = stat.clone()

    for _ in range(5):
        X = np.random.random(10)
        Y = np.random.random(10)
        batch_stat.update_many(X, Y)
        for x, y in zip(X, Y):
            stat.update(x, y)

    assert math.isclose(batch_stat.get(), stat.get())



from __future__ import annotations

import math

import pytest
from sklearn import linear_model as sklm

from river import anomaly, datasets, optim

tests = {
    "Vanilla": (
        {"optimizer": optim.SGD(1e-2), "nu": 0.5},
        {"learning_rate": "constant", "eta0": 1e-2, "nu": 0.5},
    ),
    "No intercept": (
        {"optimizer": optim.SGD(1e-2), "nu": 0.5, "intercept_lr": 0.0},
        {"learning_rate": "constant", "eta0": 1e-2, "nu": 0.5, "fit_intercept": False},
    ),
}


@pytest.mark.parametrize(
    "river_params, sklearn_params",
    tests.values(),
    ids=tests.keys(),
)
def test_sklearn_coherence(river_params, sklearn_params):
    """Checks that the sklearn and river implementations produce the same results."""

    rv = anomaly.OneClassSVM(**river_params)
    sk = sklm.SGDOneClassSVM(**sklearn_params)

    for x, _ in datasets.Phishing().take(100):
        rv.learn_one(x)
        sk.partial_fit([list(x.values())])

    for i, w in enumerate(rv.weights.values()):
        assert math.isclose(w, sk.coef_[i])



from __future__ import annotations


def test_issue_1272():
    """

    https://github.com/online-ml/river/issues/1272

    >>> import river
    >>> from river import forest, metrics

    >>> model = forest.ARFClassifier(metric=metrics.CrossEntropy())
    >>> model.learn_one({"x": 1}, True)

    >>> model = forest.ARFClassifier()
    >>> model.learn_one({"x": 1}, True)

    """



from __future__ import annotations

import pandas as pd

from river import compose, linear_model, preprocessing


def test_pipeline_funcs():
    def a(x):
        pass

    def b(x):
        pass

    pipelines = [
        compose.FuncTransformer(a) | b,
        compose.FuncTransformer(a) | ("b", b),
        compose.FuncTransformer(a) | ("b", compose.FuncTransformer(b)),
        a | compose.FuncTransformer(b),
        ("a", a) | compose.FuncTransformer(b),
        ("a", compose.FuncTransformer(a)) | compose.FuncTransformer(b),
    ]

    for pipeline in pipelines:
        assert str(pipeline) == "a | b"


def test_pipeline_add_at_start():
    def a(x):
        pass

    pipeline = preprocessing.StandardScaler() | linear_model.LinearRegression()
    pipeline = a | pipeline
    assert str(pipeline) == "a | StandardScaler | LinearRegression"


def test_union_funcs():
    def a(x):
        pass

    def b(x):
        pass

    pipelines = [
        compose.FuncTransformer(a) + b,
        compose.FuncTransformer(a) + ("b", b),
        compose.FuncTransformer(a) + ("b", compose.FuncTransformer(b)),
        a + compose.FuncTransformer(b),
        ("a", a) + compose.FuncTransformer(b),
        ("a", compose.FuncTransformer(a)) + compose.FuncTransformer(b),
    ]

    for i, pipeline in enumerate(pipelines):
        print(i, str(pipeline))
        assert str(pipeline) == "a + b"


def test_learn_one_with_learn_during_predict():
    pipeline = compose.Pipeline(
        ("scale", preprocessing.StandardScaler()),
        ("log_reg", linear_model.LogisticRegression()),
    )

    dataset = [(dict(a=x, b=x), bool(x % 2)) for x in range(100)]

    for x, y in dataset:
        counts_pre = dict(pipeline["scale"].counts)
        with compose.learn_during_predict():
            pipeline.learn_one(x, y)
        counts_no_learn = dict(pipeline["scale"].counts)
        pipeline.learn_one(x, y)
        counts_post = dict(pipeline["scale"].counts)

        assert counts_pre != counts_post
        assert counts_pre == counts_no_learn


def test_learn_many_with_learn_during_predict():
    pipeline = compose.Pipeline(
        ("scale", preprocessing.StandardScaler()),
        ("log_reg", linear_model.LogisticRegression()),
    )

    dataset = [(dict(a=x, b=x), x) for x in range(100)]

    for i in range(0, len(dataset), 5):
        X = pd.DataFrame([x for x, _ in dataset][i : i + 5])
        y = pd.Series([bool(y % 2) for _, y in dataset][i : i + 5])

        counts_pre = dict(pipeline["scale"].counts)
        with compose.learn_during_predict():
            pipeline.learn_many(X, y)
        counts_no_learn = dict(pipeline["scale"].counts)
        pipeline.learn_many(X, y)
        counts_post = dict(pipeline["scale"].counts)

        assert counts_pre != counts_post
        assert counts_pre == counts_no_learn


def test_list_of_funcs():
    def f(x):
        return {"f": 1}

    def g(x):
        return {"g": 2}

    def times_2(x):
        return {k: v * 2 for k, v in x.items()}

    expected = {"f": 2, "g": 4}
    assert compose.Pipeline([f, g], times_2).transform_one(None) == expected
    assert ([f, g] | compose.FuncTransformer(times_2)).transform_one(None) == expected


def test_get():
    model = preprocessing.StandardScaler() | linear_model.LinearRegression()

    assert isinstance(model["StandardScaler"], preprocessing.StandardScaler)
    assert isinstance(model["LinearRegression"], linear_model.LinearRegression)
    assert isinstance(model[0], preprocessing.StandardScaler)
    assert isinstance(model[1], linear_model.LinearRegression)
    assert isinstance(model[-1], linear_model.LinearRegression)
    assert isinstance(model[-2], preprocessing.StandardScaler)



from __future__ import annotations

import datetime as dt

import pytest

from river import proba, stats, utils


def test_with_counter():
    """
    >>> from river import utils
    >>> import collections
    >>> collections.Counter.revert = collections.Counter.subtract

    >>> counter = utils.Rolling(collections.Counter(), window_size=3)

    >>> for i in range(5):
    ...     counter.update([i])

    >>> counter
    Counter({2: 1, 3: 1, 4: 1, 0: 0, 1: 0})

    >>> counter.most_common(3)
    [(2, 1), (3, 1), (4, 1)]

    >>> counter[4]
    1

    """


def test_rolling_with_not_rollable():
    with pytest.raises(ValueError):
        utils.Rolling(stats.Quantile(), window_size=10)


def test_time_rolling_with_not_rollable():
    with pytest.raises(ValueError):
        utils.TimeRolling(stats.Quantile(), period=dt.timedelta(seconds=10))


def test_issue_1343():
    """

    https://github.com/online-ml/river/issues/1343

    """
    rmean = utils.TimeRolling(proba.MultivariateGaussian(), period=dt.timedelta(microseconds=1))
    t = dt.datetime.now()
    rmean.update({"a": 0}, t=t)
    rmean.update({"a": 1}, t=t)



from __future__ import annotations

import pytest
from sklearn.datasets import make_blobs

from river import metrics, stream, utils
from river.cluster import DBSTREAM


def build_dbstream(fading_factor=0.01, intersection_factor=0.05):
    return DBSTREAM(
        fading_factor=fading_factor,
        clustering_threshold=1,
        cleanup_interval=1,
        intersection_factor=intersection_factor,
    )


def add_cluster(dbstream, initial_point, move_towards, times=1):
    dbstream.learn_one(initial_point)
    for _ in range(times):
        dbstream.learn_one(move_towards)


def assert_micro_cluster_properties(cluster, center, last_update=None):
    assert cluster.center == pytest.approx(center)
    if last_update is not None:
        assert cluster.last_update == last_update


def test_cluster_formation_and_cleanup():
    dbstream = build_dbstream()

    X = [
        {1: 1},
        {1: 2},
        {1: 3},
        {1: 3},
        {1: 3},
        {1: 5},
        {1: 7},
        {1: 9},
        {1: 10},
        {1: 11},
        {1: 11},
        {1: 12},
        {1: 13},
        {1: 11},
        {1: 15},
        {1: 15},
        {1: 16},
        {1: 17},
        {1: 17},
        {1: 17},
    ]

    for x in X:
        dbstream.learn_one(x)

    assert len(dbstream._micro_clusters) == 4
    assert_micro_cluster_properties(dbstream.micro_clusters[2], center={1: 3}, last_update=4)
    assert_micro_cluster_properties(dbstream.micro_clusters[7], center={1: 11}, last_update=13)
    assert_micro_cluster_properties(dbstream.micro_clusters[8], center={1: 15}, last_update=15)
    assert_micro_cluster_properties(dbstream.micro_clusters[10], center={1: 17}, last_update=19)

    assert dbstream.predict_one({1: 2.0}) == 0
    assert dbstream.predict_one({1: 13.0}) == 1
    assert dbstream.predict_one({1: 13 + 1e-10}) == 2
    assert dbstream.predict_one({1: 16 - 1e-10}) == 2
    assert dbstream.predict_one({1: 18}) == 3

    assert len(dbstream._clusters) == 4
    assert dbstream.s == dbstream.s_t == {}


def test_with_two_micro_clusters():
    dbstream = build_dbstream()

    add_cluster(dbstream, initial_point={1: 1, 2: 1}, move_towards={1: 1.7, 2: 1.7}, times=25)
    add_cluster(dbstream, initial_point={1: 3, 2: 3}, move_towards={1: 2.3, 2: 2.3}, times=25)

    assert len(dbstream.micro_clusters) == 2
    assert_micro_cluster_properties(
        dbstream.micro_clusters[0], center={1: 2.137623, 2: 2.137623}, last_update=51
    )
    assert_micro_cluster_properties(
        dbstream.micro_clusters[1], center={1: 2.914910, 2: 2.914910}, last_update=51
    )

    assert dbstream.s == {0: {1: 23.033438964246173}}
    assert dbstream.s_t == {0: {1: 51}}

    dbstream._recluster()
    assert len(dbstream.clusters) == 1
    assert_micro_cluster_properties(dbstream.clusters[0], center={1: 2.415239, 2: 2.415239})


def test_density_graph_with_three_micro_clusters():
    dbstream = build_dbstream()

    add_cluster(dbstream, initial_point={1: 1, 2: 1}, move_towards={1: 1.7, 2: 1.7}, times=25)
    add_cluster(dbstream, initial_point={1: 3, 2: 3}, move_towards={1: 2.3, 2: 2.3}, times=25)
    # Points in the middle of first and second micro-clusters
    for _ in range(5):
        dbstream.learn_one({1: 2, 2: 2})

    assert dbstream.s == {0: {1: 23.033438964246173}}
    assert dbstream.s_t == {0: {1: 51}}

    add_cluster(dbstream, initial_point={1: 4, 2: 4}, move_towards={1: 3.3, 2: 3.3}, times=25)
    # Points in the middle of second and third micro-clusters
    for _ in range(4):
        dbstream.learn_one({1: 3, 2: 3})

    assert len(dbstream._micro_clusters) == 3
    assert_micro_cluster_properties(
        dbstream.micro_clusters[0], center={1: 2.0, 2: 2.0}, last_update=56
    )
    assert_micro_cluster_properties(
        dbstream.micro_clusters[1], center={1: 3.0, 2: 3.0}, last_update=86
    )
    assert_micro_cluster_properties(
        dbstream.micro_clusters[2], center={1: 3.982141, 2: 3.982141}, last_update=82
    )

    assert dbstream.s[0] == pytest.approx({1: 23.033439})
    assert dbstream.s[1] == pytest.approx({2: 23.033439})
    assert dbstream.s_t == {0: {1: 51}, 1: {2: 82}}

    dbstream._recluster()
    assert len(dbstream.clusters) == 1
    print(dbstream.clusters[0].center)
    assert_micro_cluster_properties(dbstream.clusters[0], center={1: 2.800788, 2: 2.800788})


def test_density_graph_with_removed_microcluster():
    dbstream = build_dbstream(fading_factor=0.1, intersection_factor=0.3)

    add_cluster(dbstream, initial_point={1: 1, 2: 1}, move_towards={1: 1.7, 2: 1.7}, times=25)
    add_cluster(dbstream, initial_point={1: 3, 2: 3}, move_towards={1: 2.3, 2: 2.3}, times=25)
    # Points in the middle of first and second micro-clusters
    for _ in range(5):
        dbstream.learn_one({1: 2, 2: 2})

    add_cluster(dbstream, initial_point={1: 3.5, 2: 3.5}, move_towards={1: 2.9, 2: 2.9}, times=25)

    # Points in the middle of second and third micro-clusters
    for _ in range(4):
        dbstream.learn_one({1: 2.6, 2: 2.6})

    assert len(dbstream._micro_clusters) == 2
    assert_micro_cluster_properties(
        dbstream.micro_clusters[0], center={1: 2.023498, 2: 2.023498}, last_update=86
    )
    assert_micro_cluster_properties(
        dbstream.micro_clusters[1], center={1: 2.766543, 2: 2.766543}, last_update=86
    )

    assert dbstream.s == {0: {1: 4.702391097045977}}
    assert dbstream.s_t == {0: {1: 86}}

    dbstream._recluster()
    assert len(dbstream.clusters) == 1
    assert_micro_cluster_properties(dbstream.clusters[0], center={1: 2.560647, 2: 2.560647})


def test_dbstream_synthetic_sklearn():
    centers = [(-10, -10), (-5, -5), (0, 0), (5, 5), (10, 10)]
    cluster_std = [0.6] * 5

    # Create a dataset with 15000 data points with 5 centers and cluster SD of 0.6 each
    X, y = make_blobs(
        n_samples=15_000, cluster_std=cluster_std, centers=centers, n_features=2, random_state=42
    )

    dbstream = DBSTREAM(
        clustering_threshold=2,
        fading_factor=0.05,
        intersection_factor=0.1,
        cleanup_interval=1.0,
        minimum_weight=1.0,
    )

    # Use VBeta as the metric to investigate the performance of DBSTREAM
    v_beta = metrics.VBeta(beta=1.0)

    for x, y_true in stream.iter_array(X, y):
        dbstream.learn_one(x)
        y_pred = dbstream.predict_one(x)
        v_beta.update(y_true, y_pred)

    assert len(dbstream._micro_clusters) == 12
    assert round(v_beta.get(), 4) == 0.9816

    assert dbstream.s.keys() == dbstream.s_t.keys()

    dbstream._recluster()

    # Check that the resulted cluster centers are close to the expected centers
    dbstream_expected_centers = {
        0: {0: 10, 1: 10},
        1: {0: -5, 1: -5},
        2: {0: 0, 1: 0},
        3: {0: 5, 1: 5},
        4: {0: -10, 1: -10},
    }

    for i in dbstream.centers.keys():
        assert (
            utils.math.minkowski_distance(dbstream.centers[i], dbstream_expected_centers[i], 2)
            < 0.2
        )



from __future__ import annotations

import random

import pytest

from river import drift, tree
from river.datasets import synth


def get_classification_data():
    return synth.LED(seed=42).take(500)


def get_regression_data():
    return synth.Friedman(seed=42).take(500)


@pytest.mark.parametrize(
    "dataset, model",
    [
        (
            get_classification_data(),
            tree.HoeffdingTreeClassifier(
                leaf_prediction="mc",
                max_size=0.05,
                grace_period=50,
                memory_estimate_period=50,
                splitter=tree.splitter.ExhaustiveSplitter(),
            ),
        ),
        (
            get_classification_data(),
            tree.HoeffdingAdaptiveTreeClassifier(
                leaf_prediction="mc",
                max_size=0.05,
                grace_period=50,
                memory_estimate_period=50,
                splitter=tree.splitter.ExhaustiveSplitter(),
            ),
        ),
        (
            get_classification_data(),
            tree.ExtremelyFastDecisionTreeClassifier(
                leaf_prediction="mc",
                max_size=0.05,
                grace_period=50,
                memory_estimate_period=50,
                splitter=tree.splitter.ExhaustiveSplitter(),
            ),
        ),
    ],
)
def test_memory_usage_class(dataset, model):
    for x, y in dataset:
        model.learn_one(x, y)

    assert model._raw_memory_usage / (2**20) < 0.065


@pytest.mark.parametrize(
    "dataset, model",
    [
        (
            get_regression_data(),
            tree.HoeffdingTreeRegressor(
                leaf_prediction="mean", max_size=0.5, memory_estimate_period=100
            ),
        ),
        (
            get_regression_data(),
            tree.HoeffdingAdaptiveTreeRegressor(
                leaf_prediction="mean", max_size=0.5, memory_estimate_period=100
            ),
        ),
    ],
)
def test_memory_usage_reg(dataset, model):
    for x, y in dataset:
        model.learn_one(x, y)

    assert model._raw_memory_usage / (2**20) < 0.5


def test_memory_usage_multitarget():
    dataset = get_regression_data()

    model = tree.iSOUPTreeRegressor(
        leaf_prediction="mean",
        max_size=0.5,
        memory_estimate_period=100,
    )

    for x, y in dataset:
        # Handcrafted targets
        y_ = {0: y, 1: 2 * y, 2: 3 * y}
        model.learn_one(x, y_)

    assert model._raw_memory_usage / (2**20) < 0.5


def test_efdt_split_reevaluation():
    dataset = synth.SEA(seed=7, variant=2).take(500)

    model = tree.ExtremelyFastDecisionTreeClassifier(
        leaf_prediction="nb",
        grace_period=50,
        min_samples_reevaluate=10,
        split_criterion="hellinger",
        delta=0.1,
    )

    max_depth = -1
    for x, y in dataset:
        model.learn_one(x, y)

        if model.height > max_depth:
            max_depth = model.height

    assert model.height != max_depth


def test_drift_adaptation_hatc():
    rng = random.Random(42)
    dataset = iter(synth.Sine(seed=8, classification_function=0, has_noise=True))

    model = tree.HoeffdingAdaptiveTreeClassifier(
        leaf_prediction="mc",
        grace_period=10,
        drift_detector=drift.ADWIN(0.1),
        delta=0.1,
        drift_window_threshold=2,
        seed=42,
        max_depth=3,
    )

    for i in range(1000):
        if i % 200 == 0 and i > 0:
            dataset = iter(
                synth.Sine(seed=8, classification_function=rng.randint(0, 3), has_noise=False)
            )

        x, y = next(dataset)
        model.learn_one(x, y)

    assert model._n_switch_alternate_trees > 0


def test_drift_adaptation_hatr():
    dataset = synth.Friedman(seed=7).take(1000)

    model = tree.HoeffdingAdaptiveTreeRegressor(
        leaf_prediction="mean",
        grace_period=10,
        delta=0.1,
        drift_detector=drift.ADWIN(0.1),
        drift_window_threshold=10,
        seed=7,
        max_depth=3,
    )

    for i, (x, y) in enumerate(dataset):
        y_ = y
        if i > 500:
            # Emulate an abrupt drift
            y_ = 1.5 * y
        model.learn_one(x, y_)

    assert model._n_alternate_trees > 0



"""General tests that all estimators need to pass."""

from __future__ import annotations

import importlib
import inspect

import pytest
from sklearn import linear_model as sk_linear_model

from river import (
    anomaly,
    base,
    checks,
    compat,
    compose,
    facto,
    feature_extraction,
    feature_selection,
    imblearn,
    linear_model,
    model_selection,
    multiclass,
    neighbors,
    neural_net,
    preprocessing,
    time_series,
)
from river.compat.river_to_sklearn import River2SKLBase
from river.compat.sklearn_to_river import SKL2RiverBase


def iter_estimators():
    for submodule in importlib.import_module("river.api").__all__:

        def is_estimator(obj):
            return inspect.isclass(obj) and issubclass(obj, base.Estimator)

        for _, obj in inspect.getmembers(
            importlib.import_module(f"river.{submodule}"), is_estimator
        ):
            yield obj


def iter_estimators_which_can_be_tested():
    ignored = (
        River2SKLBase,
        SKL2RiverBase,
        anomaly.LocalOutlierFactor,  # needs warm-start to work correctly
        compose.FuncTransformer,
        compose.Grouper,
        compose.Pipeline,
        compose.Prefixer,
        compose.Renamer,
        compose.Suffixer,
        compose.TargetTransformRegressor,
        facto.FFMClassifier,
        facto.FFMRegressor,
        facto.FMClassifier,
        facto.FMRegressor,
        facto.FwFMClassifier,
        facto.FwFMRegressor,
        facto.HOFMClassifier,
        facto.HOFMRegressor,
        feature_extraction.Agg,
        feature_extraction.TargetAgg,
        feature_selection.PoissonInclusion,
        imblearn.RandomOverSampler,
        imblearn.RandomUnderSampler,
        imblearn.RandomSampler,
        model_selection.SuccessiveHalvingClassifier,
        neighbors.LazySearch,
        neural_net.MLPRegressor,
        preprocessing.PreviousImputer,
        preprocessing.OneHotEncoder,
        preprocessing.StatImputer,
        time_series.base.Forecaster,
    )

    def can_be_tested(estimator):
        return not inspect.isabstract(estimator) and not issubclass(estimator, ignored)

    for estimator in filter(can_be_tested, iter_estimators()):
        for params in estimator._unit_test_params():
            yield estimator(**params)


@pytest.mark.parametrize(
    "estimator, check",
    [
        pytest.param(
            estimator,
            check,
            id=f"{estimator}:{check.__name__}",
        )
        for estimator in list(iter_estimators_which_can_be_tested())
        + [
            preprocessing.StandardScaler() | linear_model.LinearRegression(),
            preprocessing.StandardScaler() | linear_model.PAClassifier(),
            (
                preprocessing.StandardScaler()
                | preprocessing.TargetStandardScaler(
                    regressor=linear_model.LinearRegression(),
                )
            ),
            (
                preprocessing.StandardScaler()
                | multiclass.OneVsRestClassifier(linear_model.LogisticRegression())
            ),
            (
                preprocessing.StandardScaler()
                | multiclass.OneVsRestClassifier(linear_model.PAClassifier())
            ),
            preprocessing.MinMaxScaler() + preprocessing.StandardScaler(),
            (
                feature_extraction.PolynomialExtender()
                | preprocessing.StandardScaler()
                | linear_model.LinearRegression()
            ),
            preprocessing.MinMaxScaler() | anomaly.HalfSpaceTrees(),
            (
                preprocessing.StandardScaler()
                | compat.convert_sklearn_to_river(sk_linear_model.SGDRegressor(tol=1e-10))
            ),
        ]
        for check in checks.yield_checks(estimator)
        if check.__name__ not in estimator._unit_test_skips()
    ],
)
def test_check_estimator(estimator, check):
    check(estimator.clone())



from __future__ import annotations

import numpy as np

from river import datasets, preprocessing


def test_gaussian_random_projector_dot_product():
    dataset = datasets.TrumpApproval()
    projector = preprocessing.GaussianRandomProjector(n_components=3)

    for x, y in dataset:
        y = projector.transform_one(x)
        y_arr = np.array(list(y.values()))
        x_arr = np.array(list(x.values()))
        P = np.array(
            [[projector._projection_matrix[i, j] for j in x] for i in range(projector.n_components)]
        )
        np.testing.assert_allclose(x_arr @ P.T, y_arr)


def test_sparse_random_projector_dot_product():
    dataset = datasets.TrumpApproval()
    projector = preprocessing.SparseRandomProjector(n_components=3, density=0.5)

    for x, y in dataset:
        y = projector.transform_one(x)
        y_arr = np.array(list(y.values()))
        x_arr = np.array(list(x.values()))
        P = np.array(
            [
                [projector._projection_matrix[j].get(i, 0) for j in x]
                for i in range(projector.n_components)
            ]
        )
        np.testing.assert_allclose(x_arr @ P.T, y_arr)


def test_sparse_random_projector_size():
    dataset = datasets.TrumpApproval()
    projector = preprocessing.SparseRandomProjector(n_components=3, density=0.5)

    for x, y in dataset:
        projector.transform_one(x)
        break

    n_weights = sum(len(v) for v in projector._projection_matrix.values())
    assert n_weights < len(x) * projector.n_components



from __future__ import annotations

import json
import os
import zipfile

import pytest
import sqlalchemy as sql

from river import stream


@pytest.fixture
def pokedb():
    engine = sql.create_engine("sqlite://")  # in-memory

    # Load the fixtures
    here = os.path.dirname(os.path.realpath(__file__))
    with zipfile.ZipFile(os.path.join(here, "pokedb.zip")) as z:
        pokemons = json.loads(z.read("pokemons.json"))
        types = json.loads(z.read("types.json"))
        pokemon_types = json.loads(z.read("pokemon_types.json"))

    # Define the tables
    metadata = sql.MetaData()

    t_pokemons = sql.Table(
        "pokemons",
        metadata,
        sql.Column("id", sql.Integer, primary_key=True),
        sql.Column("name", sql.String),
        sql.Column("HP", sql.Integer),
        sql.Column("Attack", sql.Integer),
        sql.Column("Defense", sql.Integer),
        sql.Column("Sp. Attack", sql.Integer),
        sql.Column("Sp. Defense", sql.Integer),
        sql.Column("Speed", sql.Integer),
    )

    t_types = sql.Table(
        "types",
        metadata,
        sql.Column("id", sql.Integer, primary_key=True),
        sql.Column("name", sql.String),
    )

    t_pokemon_types = sql.Table(
        "pokemon_types",
        metadata,
        sql.Column("pokemon_id", sql.Integer, primary_key=True),
        sql.Column("type_id", sql.Integer, primary_key=True),
        sql.Column("no", sql.Integer, primary_key=True),
    )

    # Create the tables
    metadata.create_all(engine)

    # Insert the fixtures
    with engine.connect() as conn:
        conn.execute(t_pokemons.insert(), pokemons)
        conn.execute(t_types.insert(), types)
        conn.execute(t_pokemon_types.insert(), pokemon_types)
        conn.commit()

    return engine


def test_iter_sql(pokedb):
    with pokedb.connect() as conn:
        dataset = stream.iter_sql(query="SELECT * FROM pokemons;", conn=conn)
        x, y = next(dataset)
        assert x["name"] == "Bulbasaur"
        assert y is None

    # This raises an exception because the resource is closed...
    with pytest.raises(sql.exc.ResourceClosedError):
        for x, y in stream.iter_sql(query="SELECT * FROM pokemons;", conn=conn):
            pass

    # ... and yet we can still stream over the results because SQLAlchemy prefetches them
    x, y = next(dataset)
    assert x["name"] == "Ivysaur"

    # The Pokedex from generation 1 contains 151 pokemons, and we've already seen 2 of them
    assert sum(1 for _ in dataset) == 149

    # Check that the stream is depleted
    assert sum(1 for _ in dataset) == 0


def test_iter_sql_join(pokedb):
    query = """
        SELECT
            p.name,
            t1.name AS type_1,
            t2.name AS type_2
        FROM
            pokemons p,
            pokemon_types pt1,
            pokemon_types pt2,
            types t1,
            types t2
        WHERE
            pt1.no = 1 AND
            pt1.pokemon_id = p.id AND
            pt1.type_id = t1.id AND

            pt2.no = 2 AND
            pt2.pokemon_id = p.id AND
            pt2.type_id = t2.id;
    """

    with pokedb.connect() as conn:
        dataset = stream.iter_sql(query=query, conn=conn)
        x, y = next(dataset)
        assert x["name"] == "Bulbasaur"
        assert x["type_1"] == "Grass"
        assert x["type_2"] == "Poison"



from __future__ import annotations

import pytest

from river import compose, linear_model, optim, preprocessing, tree, utils


@pytest.mark.parametrize(
    "model, param_grid, count",
    [
        (
            linear_model.LinearRegression(),
            {
                "optimizer": [
                    (optim.SGD, {"lr": [1, 2]}),
                    (
                        optim.Adam,
                        {
                            "beta_1": [0.1, 0.01, 0.001],
                            "lr": [0.1, 0.01, 0.001, 0.0001],
                        },
                    ),
                ]
            },
            2 + 3 * 4,
        ),
        (
            preprocessing.StandardScaler() | linear_model.LinearRegression(),
            {
                "LinearRegression": {
                    "optimizer": [
                        (optim.SGD, {"lr": [1, 2]}),
                        (
                            optim.Adam,
                            {
                                "beta_1": [0.1, 0.01, 0.001],
                                "lr": [0.1, 0.01, 0.001, 0.0001],
                            },
                        ),
                    ]
                }
            },
            2 + 3 * 4,
        ),
        (
            compose.Pipeline(("Scaler", None), linear_model.LinearRegression()),
            {
                "Scaler": [
                    preprocessing.MinMaxScaler(),
                    preprocessing.MaxAbsScaler(),
                    preprocessing.StandardScaler(),
                ],
                "LinearRegression": {"optimizer": {"lr": [1e-1, 1e-2, 1e-3]}},
            },
            3 * 3,
        ),
    ],
)
def test_expand_param_grid_count(model, param_grid, count):
    assert len(utils.expand_param_grid(model, param_grid)) == count


def test_decision_tree_max_depth():
    model = tree.HoeffdingTreeClassifier()

    max_depths = [1, 2, 3, 4, 5, 6]
    models = utils.expand_param_grid(model, {"max_depth": max_depths})

    for model, max_depth in zip(models, max_depths):
        assert model.max_depth == max_depth



from __future__ import annotations

import pytest

from river import feature_extraction


@pytest.mark.parametrize(
    "params, text, expected_ngrams",
    [
        pytest.param(
            *case,
            id=f"#{i}",
        )
        for i, case in enumerate(
            [
                ({}, "one two three", ["one", "two", "three"]),
                (
                    {},
                    """one   two\tthree four\t\tfive
            six

            seven""",
                    ["one", "two", "three", "four", "five", "six", "seven"],
                ),
                (
                    {"ngram_range": (1, 2)},
                    "one two three",
                    ["one", "two", "three", ("one", "two"), ("two", "three")],
                ),
                ({"ngram_range": (2, 2)}, "one two three", [("one", "two"), ("two", "three")]),
                (
                    {"ngram_range": (2, 3)},
                    "one two three",
                    [("one", "two"), ("two", "three"), ("one", "two", "three")],
                ),
                ({"stop_words": {"two", "three"}}, "one two three four", ["one", "four"]),
                (
                    {"stop_words": {"two", "three"}, "ngram_range": (1, 2)},
                    "one two three four",
                    ["one", "four", ("one", "four")],
                ),
            ]
        )
    ],
)
def test_ngrams(params, text, expected_ngrams):
    bow = feature_extraction.BagOfWords(**params)
    ngrams = list(bow.process_text(text))
    assert expected_ngrams == ngrams
